@@ -11,9 +11,60 @@
 #include <linux/slab.h>
 #include <linux/bitmap.h>
 #include <linux/scatterlist.h>
+#include <net/ip_tunnels.h>
 #include <net/xfrm.h>
 #include <crypto/algapi.h>
 
+struct encryption_skb_cb {
+	uint8_t ds;
+	uint8_t num_frags;
+	unsigned int plaintext_len, trailer_len;
+	struct sk_buff *trailer;
+	uint64_t nonce;
+};
+
+struct encryption_ctx {
+	struct padata_priv padata;
+	struct sk_buff_head queue;
+	void (*callback)(struct sk_buff_head *, struct wireguard_peer *);
+	struct wireguard_peer *peer;
+	struct noise_keypair *keypair;
+};
+
+struct decryption_ctx {
+	struct padata_priv padata;
+	struct sk_buff *skb;
+	void (*callback)(struct sk_buff *skb, struct wireguard_peer *, struct sockaddr_storage *, bool used_new_key, int err);
+	struct noise_keypair *keypair;
+	struct sockaddr_storage addr;
+	uint64_t nonce;
+	uint8_t num_frags;
+	int ret;
+};
+
+static struct kmem_cache *encryption_ctx_cache;
+static struct kmem_cache *decryption_ctx_cache;
+
+int packet_init_data_caches(void)
+{
+	BUILD_BUG_ON(sizeof(struct encryption_skb_cb) > sizeof(((struct sk_buff *)0)->cb));
+	encryption_ctx_cache = kmem_cache_create("wireguard_encryption_ctx", sizeof(struct encryption_ctx), 0, 0, NULL);
+	if (!encryption_ctx_cache)
+		return -ENOMEM;
+	decryption_ctx_cache = kmem_cache_create("wireguard_decryption_ctx", sizeof(struct decryption_ctx), 0, 0, NULL);
+	if (!decryption_ctx_cache) {
+		kmem_cache_destroy(encryption_ctx_cache);
+		return -ENOMEM;
+	}
+	return 0;
+}
+
+void packet_deinit_data_caches(void)
+{
+	kmem_cache_destroy(encryption_ctx_cache);
+	kmem_cache_destroy(decryption_ctx_cache);
+}
+
 /* This is RFC6479, a replay detection bitmap algorithm that avoids bitshifts */
 static inline bool counter_validate(union noise_counter *counter, u64 their_counter)
 {
@@ -48,13 +99,13 @@ out:
 }
 #include "selftest/counter.h"
 
-static inline size_t skb_padding(struct sk_buff *skb)
+static inline unsigned int skb_padding(struct sk_buff *skb)
 {
 	/* We do this modulo business with the MTU, just in case the networking layer
 	 * gives us a packet that's bigger than the MTU. Now that we support GSO, this
 	 * shouldn't be a real problem, and this can likely be removed. But, caution! */
-	size_t last_unit = skb->len % skb->dev->mtu;
-	size_t padded_size = (last_unit + MESSAGE_PADDING_MULTIPLE - 1) & ~(MESSAGE_PADDING_MULTIPLE - 1);
+	unsigned int last_unit = skb->len % skb->dev->mtu;
+	unsigned int padded_size = (last_unit + MESSAGE_PADDING_MULTIPLE - 1) & ~(MESSAGE_PADDING_MULTIPLE - 1);
 	if (padded_size > skb->dev->mtu)
 		padded_size = skb->dev->mtu;
 	return padded_size - last_unit;
@@ -75,9 +126,10 @@ static inline void skb_reset(struct sk_buff *skb)
 	skb_probe_transport_header(skb, 0);
 }
 
-static inline void skb_encrypt(struct sk_buff *skb, struct packet_data_encryption_ctx *ctx)
+static inline void skb_encrypt(struct sk_buff *skb, struct noise_keypair *keypair, bool have_simd)
 {
-	struct scatterlist sg[ctx->num_frags]; /* This should be bound to at most 128 by the caller. */
+	struct encryption_skb_cb *cb = (struct encryption_skb_cb *)skb->cb;
+	struct scatterlist sg[cb->num_frags]; /* This should be bound to at most 128 by the caller. */
 	struct message_data *header;
 
 	/* We have to remember to add the checksum to the innerpacket, in case the receiver forwards it. */
@@ -87,20 +139,17 @@ static inline void skb_encrypt(struct sk_buff *skb, struct packet_data_encryptio
 	/* Only after checksumming can we safely add on the padding at the end and the header. */
 	header = (struct message_data *)skb_push(skb, sizeof(struct message_data));
 	header->header.type = MESSAGE_DATA;
-	header->key_idx = ctx->keypair->remote_index;
-	header->counter = cpu_to_le64(ctx->nonce);
-	pskb_put(skb, ctx->trailer, ctx->trailer_len);
+	header->key_idx = keypair->remote_index;
+	header->counter = cpu_to_le64(cb->nonce);
+	pskb_put(skb, cb->trailer, cb->trailer_len);
 
 	/* Now we can encrypt the scattergather segments */
-	sg_init_table(sg, ctx->num_frags);
-	skb_to_sgvec(skb, sg, sizeof(struct message_data), noise_encrypted_len(ctx->plaintext_len));
-	chacha20poly1305_encrypt_sg(sg, sg, ctx->plaintext_len, NULL, 0, ctx->nonce, ctx->keypair->sending.key);
-
-	/* When we're done, we free the reference to the key pair */
-	noise_keypair_put(ctx->keypair);
+	sg_init_table(sg, cb->num_frags);
+	skb_to_sgvec(skb, sg, sizeof(struct message_data), noise_encrypted_len(cb->plaintext_len));
+	chacha20poly1305_encrypt_sg(sg, sg, cb->plaintext_len, NULL, 0, cb->nonce, keypair->sending.key, have_simd);
 }
 
-static inline bool skb_decrypt(struct sk_buff *skb, unsigned int num_frags, uint64_t nonce, struct noise_symmetric_key *key)
+static inline bool skb_decrypt(struct sk_buff *skb, uint8_t num_frags, uint64_t nonce, struct noise_symmetric_key *key)
 {
 	struct scatterlist sg[num_frags]; /* This should be bound to at most 128 by the caller. */
 
@@ -140,22 +189,34 @@ static inline bool get_encryption_nonce(uint64_t *nonce, struct noise_symmetric_
 	return true;
 }
 
+static inline void queue_encrypt_reset(struct sk_buff_head *queue, struct noise_keypair *keypair)
+{
+	struct sk_buff *skb;
+	bool have_simd = chacha20poly1305_init_simd();
+	skb_queue_walk(queue, skb) {
+		skb_encrypt(skb, keypair, have_simd);
+		skb_reset(skb);
+	}
+	chacha20poly1305_deinit_simd(have_simd);
+	noise_keypair_put(keypair);
+}
+
 #ifdef CONFIG_WIREGUARD_PARALLEL
 static void do_encryption(struct padata_priv *padata)
 {
-	struct packet_data_encryption_ctx *ctx = container_of(padata, struct packet_data_encryption_ctx, padata);
-
-	skb_encrypt(ctx->skb, ctx);
-	skb_reset(ctx->skb);
+	struct encryption_ctx *ctx = container_of(padata, struct encryption_ctx, padata);
 
+	queue_encrypt_reset(&ctx->queue, ctx->keypair);
 	padata_do_serial(padata);
 }
 
 static void finish_encryption(struct padata_priv *padata)
 {
-	struct packet_data_encryption_ctx *ctx = container_of(padata, struct packet_data_encryption_ctx, padata);
+	struct encryption_ctx *ctx = container_of(padata, struct encryption_ctx, padata);
 
-	ctx->callback(ctx->skb, ctx->peer);
+	ctx->callback(&ctx->queue, ctx->peer);
+	peer_put(ctx->peer);
+	kmem_cache_free(encryption_ctx_cache, ctx);
 }
 
 static inline int start_encryption(struct padata_instance *padata, struct padata_priv *priv, int cb_cpu)
@@ -180,15 +241,11 @@ static inline unsigned int choose_cpu(__le32 key)
 }
 #endif
 
-int packet_create_data(struct sk_buff *skb, struct wireguard_peer *peer, void(*callback)(struct sk_buff *, struct wireguard_peer *), bool parallel)
+int packet_create_data(struct sk_buff_head *queue, struct wireguard_peer *peer, void(*callback)(struct sk_buff_head *, struct wireguard_peer *))
 {
 	int ret = -ENOKEY;
 	struct noise_keypair *keypair;
-	struct packet_data_encryption_ctx *ctx = NULL;
-	u64 nonce;
-	struct sk_buff *trailer = NULL;
-	size_t plaintext_len, padding_len, trailer_len;
-	unsigned int num_frags;
+	struct sk_buff *skb;
 
 	rcu_read_lock();
 	keypair = noise_keypair_get(rcu_dereference(peer->keypairs.current_keypair));
@@ -196,54 +253,75 @@ int packet_create_data(struct sk_buff *skb, struct wireguard_peer *peer, void(*c
 		goto err_rcu;
 	rcu_read_unlock();
 
-	if (unlikely(!get_encryption_nonce(&nonce, &keypair->sending)))
-		goto err;
+	skb_queue_walk(queue, skb) {
+		struct encryption_skb_cb *cb = (struct encryption_skb_cb *)skb->cb;
+		unsigned int padding_len, num_frags;
 
-	padding_len = skb_padding(skb);
-	trailer_len = padding_len + noise_encrypted_len(0);
-	plaintext_len = skb->len + padding_len;
+		if (unlikely(!get_encryption_nonce(&cb->nonce, &keypair->sending)))
+			goto err;
 
-	/* Expand data section to have room for padding and auth tag */
-	ret = skb_cow_data(skb, trailer_len, &trailer);
-	if (unlikely(ret < 0))
-		goto err;
-	num_frags = ret;
-	ret = -ENOMEM;
-	if (unlikely(num_frags > 128))
-		goto err;
+		padding_len = skb_padding(skb);
+		cb->trailer_len = padding_len + noise_encrypted_len(0);
+		cb->plaintext_len = skb->len + padding_len;
 
-	/* Set the padding to zeros, and make sure it and the auth tag are part of the skb */
-	memset(skb_tail_pointer(trailer), 0, padding_len);
+		/* Store the ds bit in the cb */
+		cb->ds = ip_tunnel_ecn_encap(0 /* No outer TOS: no leak. TODO: should we use flowi->tos as outer? */, ip_hdr(skb), skb);
 
-	/* Expand head section to have room for our header and the network stack's headers,
-	 * plus our key and nonce in the head. */
-	ret = skb_cow_head(skb, DATA_PACKET_HEAD_ROOM);
-	if (unlikely(ret < 0))
-		goto err;
-
-	ctx = (struct packet_data_encryption_ctx *)skb->head;
-	ctx->skb = skb;
-	ctx->callback = callback;
-	ctx->peer = peer;
-	ctx->num_frags = num_frags;
-	ctx->trailer_len = trailer_len;
-	ctx->trailer = trailer;
-	ctx->plaintext_len = plaintext_len;
-	ctx->nonce = nonce;
-	ctx->keypair = keypair;
-
-#ifdef CONFIG_WIREGUARD_PARALLEL
-	if ((parallel || padata_queue_len(peer->device->parallel_send) > 0) && cpumask_weight(cpu_online_mask) > 1) {
-		unsigned int cpu = choose_cpu(keypair->remote_index);
-		ret = start_encryption(peer->device->parallel_send, &ctx->padata, cpu);
+		/* Expand data section to have room for padding and auth tag */
+		ret = skb_cow_data(skb, cb->trailer_len, &cb->trailer);
 		if (unlikely(ret < 0))
 			goto err;
+		num_frags = ret;
+		ret = -ENOMEM;
+		if (unlikely(num_frags > 128))
+			goto err;
+		cb->num_frags = num_frags;
+
+		/* Set the padding to zeros, and make sure it and the auth tag are part of the skb */
+		memset(skb_tail_pointer(cb->trailer), 0, padding_len);
+
+		/* Expand head section to have room for our header and the network stack's headers. */
+		ret = skb_cow_head(skb, DATA_PACKET_HEAD_ROOM);
+		if (unlikely(ret < 0))
+			goto err;
+
+		/* After the first time through the loop, if we've suceeded with a legitimate nonce,
+		 * then we don't want a -ENOKEY error if subsequent nonces fail. Rather, if this
+		 * condition arises, we simply want error out hard, and drop the entire queue. This
+		 * is partially lazy programming and TODO: this could be made to only requeue the
+		 * ones that had no nonce. But I'm not sure it's worth the added complexity, given
+		 * how rarely that condition should arise. */
+		ret = -EPIPE;
+	}
+
+#ifdef CONFIG_WIREGUARD_PARALLEL
+	if ((skb_queue_len(queue) > 1 || queue->next->len > 256 || padata_queue_len(peer->device->parallel_send) > 0) && cpumask_weight(cpu_online_mask) > 1) {
+		unsigned int cpu = choose_cpu(keypair->remote_index);
+		struct encryption_ctx *ctx = kmem_cache_alloc(encryption_ctx_cache, GFP_ATOMIC);
+		if (!ctx)
+			goto serial_encrypt;
+		skb_queue_head_init(&ctx->queue);
+		skb_queue_splice_init(queue, &ctx->queue);
+		ctx->callback = callback;
+		ctx->keypair = keypair;
+		ctx->peer = peer_rcu_get(peer);
+		ret = -EBUSY;
+		if (unlikely(!ctx->peer))
+			goto err_parallel;
+		ret = start_encryption(peer->device->parallel_send, &ctx->padata, cpu);
+		if (unlikely(ret < 0)) {
+			peer_put(ctx->peer);
+err_parallel:
+			skb_queue_splice(&ctx->queue, queue);
+			kmem_cache_free(encryption_ctx_cache, ctx);
+			goto err;
+		}
 	} else
 #endif
 	{
-		skb_encrypt(skb, ctx);
-		skb_reset(skb);
-		callback(skb, peer);
+serial_encrypt:
+		queue_encrypt_reset(queue, keypair);
+		callback(queue, peer);
 	}
 	return 0;
 
@@ -255,18 +333,7 @@ err_rcu:
 	return ret;
 }
 
-struct packet_data_decryption_ctx {
-	struct padata_priv padata;
-	struct sk_buff *skb;
-	void (*callback)(struct sk_buff *skb, struct wireguard_peer *, struct sockaddr_storage *, bool used_new_key, int err);
-	struct noise_keypair *keypair;
-	struct sockaddr_storage addr;
-	uint64_t nonce;
-	unsigned int num_frags;
-	int ret;
-};
-
-static void begin_decrypt_packet(struct packet_data_decryption_ctx *ctx)
+static void begin_decrypt_packet(struct decryption_ctx *ctx)
 {
 	if (unlikely(!skb_decrypt(ctx->skb, ctx->num_frags, ctx->nonce, &ctx->keypair->receiving)))
 		goto err;
@@ -280,7 +347,7 @@ err:
 	peer_put(ctx->keypair->entry.peer);
 }
 
-static void finish_decrypt_packet(struct packet_data_decryption_ctx *ctx)
+static void finish_decrypt_packet(struct decryption_ctx *ctx)
 {
 	struct noise_keypairs *keypairs;
 	bool used_new_key = false;
@@ -311,16 +378,16 @@ err:
 #ifdef CONFIG_WIREGUARD_PARALLEL
 static void do_decryption(struct padata_priv *padata)
 {
-	struct packet_data_decryption_ctx *ctx = container_of(padata, struct packet_data_decryption_ctx, padata);
+	struct decryption_ctx *ctx = container_of(padata, struct decryption_ctx, padata);
 	begin_decrypt_packet(ctx);
 	padata_do_serial(padata);
 }
 
 static void finish_decryption(struct padata_priv *padata)
 {
-	struct packet_data_decryption_ctx *ctx = container_of(padata, struct packet_data_decryption_ctx, padata);
+	struct decryption_ctx *ctx = container_of(padata, struct decryption_ctx, padata);
 	finish_decrypt_packet(ctx);
-	kfree(ctx);
+	kmem_cache_free(decryption_ctx_cache, ctx);
 }
 
 static inline int start_decryption(struct padata_instance *padata, struct padata_priv *priv, int cb_cpu)
@@ -374,11 +441,11 @@ void packet_consume_data(struct sk_buff *skb, size_t offset, struct wireguard_de
 	rcu_read_unlock();
 #ifdef CONFIG_WIREGUARD_PARALLEL
 	if (cpumask_weight(cpu_online_mask) > 1) {
-		struct packet_data_decryption_ctx *ctx;
 		unsigned int cpu = choose_cpu(idx);
+		struct decryption_ctx *ctx;
 
 		ret = -ENOMEM;
-		ctx = kzalloc(sizeof(struct packet_data_decryption_ctx), GFP_ATOMIC);
+		ctx = kmem_cache_alloc(decryption_ctx_cache, GFP_ATOMIC);
 		if (unlikely(!ctx))
 			goto err_peer;
 
@@ -390,13 +457,13 @@ void packet_consume_data(struct sk_buff *skb, size_t offset, struct wireguard_de
 		ctx->addr = addr;
 		ret = start_decryption(wg->parallel_receive, &ctx->padata, cpu);
 		if (unlikely(ret)) {
-			kfree(ctx);
+			kmem_cache_free(decryption_ctx_cache, ctx);
 			goto err_peer;
 		}
 	} else
 #endif
 	{
-		struct packet_data_decryption_ctx ctx = {
+		struct decryption_ctx ctx = {
 			.skb = skb,
 			.keypair = keypair,
 			.callback = callback,
