@@ -13,6 +13,178 @@
 #include <net/udp_tunnel.h>
 #include <net/ipv6.h>
 
+static inline int send4(struct wireguard_device *wg, struct sk_buff *skb, struct sockaddr_in *addr, uint8_t ds, struct dst_cache *cache)
+{
+	struct flowi4 fl = {
+		.daddr = addr->sin_addr.s_addr,
+		.fl4_dport = addr->sin_port,
+		.fl4_sport = htons(wg->incoming_port),
+		.flowi4_proto = IPPROTO_UDP
+	};
+	struct rtable *rt = NULL;
+	struct sock *sock;
+	int ret = 0;
+
+	skb->next = skb->prev = NULL;
+	skb->dev = netdev_pub(wg);
+
+	rcu_read_lock();
+	sock = rcu_dereference(wg->sock4);
+
+	if (unlikely(!sock)) {
+		ret = -ENONET;
+		goto err;
+	}
+
+	if (cache)
+		rt = dst_cache_get_ip4(cache, &fl.saddr);
+
+	if (!rt) {
+		security_sk_classify_flow(sock, flowi4_to_flowi(&fl));
+		rt = ip_route_output_flow(sock_net(sock), &fl, sock);
+		if (unlikely(IS_ERR(rt))) {
+			ret = PTR_ERR(rt);
+			net_dbg_ratelimited("No route to %pISpfsc, error %d\n", addr, ret);
+			goto err;
+		} else if (unlikely(rt->dst.dev == skb->dev)) {
+			ret = -ELOOP;
+			net_dbg_ratelimited("Avoiding routing loop to %pISpfsc\n", addr);
+			goto err;
+		}
+		if (cache)
+			dst_cache_set_ip4(cache, &rt->dst, fl.saddr);
+	}
+
+	udp_tunnel_xmit_skb(rt, sock, skb,
+			    fl.saddr, fl.daddr,
+			    ds, ip4_dst_hoplimit(&rt->dst), 0,
+			    fl.fl4_sport, fl.fl4_dport,
+			    false, false);
+	goto out;
+
+err:
+	kfree_skb(skb);
+out:
+	rcu_read_unlock();
+	return ret;
+}
+
+static inline int send6(struct wireguard_device *wg, struct sk_buff *skb, struct sockaddr_in6 *addr, uint8_t ds, struct dst_cache *cache)
+{
+#if IS_ENABLED(CONFIG_IPV6)
+	struct flowi6 fl = {
+		.daddr = addr->sin6_addr,
+		.fl6_dport = addr->sin6_port,
+		.fl6_sport = htons(wg->incoming_port),
+		.flowi6_oif = addr->sin6_scope_id,
+		.flowi6_proto = IPPROTO_UDP
+		/* TODO: addr->sin6_flowinfo */
+	};
+	struct dst_entry *dst = NULL;
+	struct sock *sock;
+	int ret = 0;
+
+	skb->next = skb->prev = NULL;
+	skb->dev = netdev_pub(wg);
+
+	rcu_read_lock();
+	sock = rcu_dereference(wg->sock6);
+
+	if (unlikely(!sock)) {
+		ret = -ENONET;
+		goto err;
+	}
+
+	if (cache)
+		dst = dst_cache_get_ip6(cache, &fl.saddr);
+
+	if (!dst) {
+		security_sk_classify_flow(sock, flowi6_to_flowi(&fl));
+		ret = ipv6_stub->ipv6_dst_lookup(sock_net(sock), sock, &dst, &fl);
+		if (unlikely(ret)) {
+			net_dbg_ratelimited("No route to %pISpfsc, error %d\n", addr, ret);
+			goto err;
+		} else if (unlikely(dst->dev == skb->dev)) {
+			ret = -ELOOP;
+			net_dbg_ratelimited("Avoiding routing loop to %pISpfsc\n", addr);
+			goto err;
+		}
+		if (cache)
+			dst_cache_set_ip6(cache, dst, &fl.saddr);
+	}
+
+	udp_tunnel6_xmit_skb(dst, sock, skb, skb->dev,
+			     &fl.saddr, &fl.daddr,
+			     ds, ip6_dst_hoplimit(dst), 0,
+			     fl.fl6_sport, fl.fl6_dport,
+			     false);
+	goto out;
+
+err:
+	kfree_skb(skb);
+out:
+	rcu_read_unlock();
+	return ret;
+#else
+	return -EAFNOSUPPORT;
+#endif
+}
+
+int socket_send_skb_to_peer(struct wireguard_peer *peer, struct sk_buff *skb, uint8_t ds)
+{
+	size_t skb_len = skb->len;
+	int ret = -EAFNOSUPPORT;
+
+	read_lock_bh(&peer->endpoint_lock);
+	if (peer->endpoint_addr.ss_family == AF_INET)
+		ret = send4(peer->device, skb, (struct sockaddr_in *)&peer->endpoint_addr, ds, &peer->endpoint_cache);
+	else if (peer->endpoint_addr.ss_family == AF_INET6)
+		ret = send6(peer->device, skb, (struct sockaddr_in6 *)&peer->endpoint_addr, ds, &peer->endpoint_cache);
+	if (likely(!ret))
+		peer->tx_bytes += skb_len;
+	read_unlock_bh(&peer->endpoint_lock);
+
+	return ret;
+}
+
+int socket_send_buffer_to_peer(struct wireguard_peer *peer, void *buffer, size_t len, uint8_t ds)
+{
+	struct sk_buff *skb = alloc_skb(len + SKB_HEADER_LEN, GFP_ATOMIC);
+	if (unlikely(!skb))
+		return -ENOMEM;
+	skb_reserve(skb, SKB_HEADER_LEN);
+	memcpy(skb_put(skb, len), buffer, len);
+	return socket_send_skb_to_peer(peer, skb, ds);
+}
+
+int socket_send_buffer_as_reply_to_skb(struct wireguard_device *wg, struct sk_buff *in_skb, void *out_buffer, size_t len)
+{
+	int ret = 0;
+	struct sk_buff *skb;
+	struct sockaddr_storage addr = { 0 };
+
+	if (unlikely(!in_skb))
+		return -EINVAL;
+	ret = socket_addr_from_skb(&addr, in_skb);
+	if (unlikely(ret < 0))
+		return ret;
+
+	skb = alloc_skb(len + SKB_HEADER_LEN, GFP_ATOMIC);
+	if (unlikely(!skb))
+		return -ENOMEM;
+	skb_reserve(skb, SKB_HEADER_LEN);
+	memcpy(skb_put(skb, len), out_buffer, len);
+
+	if (addr.ss_family == AF_INET)
+		ret = send4(wg, skb, (struct sockaddr_in *)&addr, 0, NULL);
+	else if (addr.ss_family == AF_INET6)
+		ret = send6(wg, skb, (struct sockaddr_in6 *)&addr, 0, NULL);
+	else
+		ret = -EAFNOSUPPORT;
+
+	return ret;
+}
+
 int socket_addr_from_skb(struct sockaddr_storage *sockaddr, struct sk_buff *skb)
 {
 	struct iphdr *ip4;
@@ -41,122 +213,6 @@ int socket_addr_from_skb(struct sockaddr_storage *sockaddr, struct sk_buff *skb)
 	return 0;
 }
 
-static inline struct dst_entry *route(struct wireguard_device *wg, struct flowi4 *fl4, struct flowi6 *fl6, struct sockaddr_storage *addr, struct sock *sock4, struct sock *sock6)
-{
-	struct dst_entry *dst = ERR_PTR(-EAFNOSUPPORT);
-
-	if (addr->ss_family == AF_INET) {
-		struct rtable *rt;
-		struct sockaddr_in *sin4 = (struct sockaddr_in *)addr;
-
-		if (unlikely(!sock4))
-			return ERR_PTR(-ENONET);
-
-		memset(fl4, 0, sizeof(struct flowi4));
-		fl4->daddr = sin4->sin_addr.s_addr;
-		fl4->fl4_dport = sin4->sin_port;
-		fl4->fl4_sport = htons(wg->incoming_port);
-		fl4->flowi4_proto = IPPROTO_UDP;
-
-		security_sk_classify_flow(sock4, flowi4_to_flowi(fl4));
-		rt = ip_route_output_flow(sock_net(sock4), fl4, sock4);
-		if (unlikely(IS_ERR(rt)))
-			dst = ERR_PTR(PTR_ERR(rt));
-		dst = &rt->dst;
-	} else if (addr->ss_family == AF_INET6) {
-#if IS_ENABLED(CONFIG_IPV6)
-		int ret;
-		struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)addr;
-
-		if (unlikely(!sock6))
-			return ERR_PTR(-ENONET);
-
-		memset(fl6, 0, sizeof(struct flowi6));
-		fl6->daddr = sin6->sin6_addr;
-		fl6->fl6_dport = sin6->sin6_port;
-		fl6->fl6_sport = htons(wg->incoming_port);
-		fl6->flowi6_oif = sin6->sin6_scope_id;
-		fl6->flowi6_proto = IPPROTO_UDP;
-		/* TODO: addr6->sin6_flowinfo */
-
-		security_sk_classify_flow(sock6, flowi6_to_flowi(fl6));
-		ret = ipv6_stub->ipv6_dst_lookup(sock_net(sock6), sock6, &dst, fl6);
-		if (unlikely(ret))
-			dst = ERR_PTR(ret);
-#endif
-	}
-	return dst;
-}
-
-static inline int send(struct net_device *dev, struct sk_buff *skb, struct dst_entry *dst, struct flowi4 *fl4, struct flowi6 *fl6, struct sockaddr_storage *addr, struct sock *sock4, struct sock *sock6, u8 dscp)
-{
-	int ret = -EAFNOSUPPORT;
-
-	skb->next = skb->prev = NULL;
-	skb->dev = dev;
-
-	if (addr->ss_family == AF_INET) {
-		if (unlikely(!sock4)) {
-			ret = -ENONET;
-			goto err;
-		}
-		udp_tunnel_xmit_skb((struct rtable *)dst, sock4, skb,
-				    fl4->saddr, fl4->daddr,
-				    dscp, ip4_dst_hoplimit(dst), 0,
-				    fl4->fl4_sport, fl4->fl4_dport,
-				    false, false);
-		return 0;
-	} else if (addr->ss_family == AF_INET6) {
-		if (unlikely(!sock6)) {
-			ret = -ENONET;
-			goto err;
-		}
-#if IS_ENABLED(CONFIG_IPV6)
-		udp_tunnel6_xmit_skb(dst, sock6, skb, dev,
-				     &fl6->saddr, &fl6->daddr,
-				     dscp, ip6_dst_hoplimit(dst), 0,
-				     fl6->fl6_sport, fl6->fl6_dport,
-				     false);
-		return 0;
-#else
-		goto err;
-#endif
-	}
-
-err:
-	kfree_skb(skb);
-	dst_release(dst);
-	return ret;
-}
-
-static inline void __socket_set_peer_dst(struct wireguard_peer *peer)
-{
-	struct dst_entry *old_dst, *new_dst;
-	lockdep_assert_held(&peer->endpoint_lock);
-
-	old_dst = peer->endpoint_dst;
-	peer->endpoint_dst = NULL;
-	wmb();
-	if (old_dst)
-		dst_release(old_dst);
-
-	rcu_read_lock();
-	new_dst = route(peer->device, &peer->endpoint_flow.fl4, &peer->endpoint_flow.fl6, &peer->endpoint_addr, rcu_dereference(peer->device->sock4), rcu_dereference(peer->device->sock6));
-	rcu_read_unlock();
-
-	if (likely(!IS_ERR(new_dst))) {
-		peer->endpoint_dst = new_dst;
-		wmb();
-	}
-}
-
-void socket_set_peer_dst(struct wireguard_peer *peer)
-{
-	write_lock_bh(&peer->endpoint_lock);
-	__socket_set_peer_dst(peer);
-	write_unlock_bh(&peer->endpoint_lock);
-}
-
 void socket_set_peer_addr(struct wireguard_peer *peer, struct sockaddr_storage *sockaddr)
 {
 	if (sockaddr->ss_family == AF_INET) {
@@ -175,126 +231,13 @@ void socket_set_peer_addr(struct wireguard_peer *peer, struct sockaddr_storage *
 		memcpy(&peer->endpoint_addr, sockaddr, sizeof(struct sockaddr_in6));
 	} else
 		return;
-	__socket_set_peer_dst(peer);
+	dst_cache_reset(&peer->endpoint_cache);
 	write_unlock_bh(&peer->endpoint_lock);
 	return;
 out:
 	read_unlock_bh(&peer->endpoint_lock);
 }
 
-static inline struct dst_entry *peer_dst_get(struct wireguard_peer *peer)
-{
-	struct dst_entry *dst = NULL;
-	read_lock_bh(&peer->endpoint_lock);
-
-	if (!peer->endpoint_dst || (peer->endpoint_dst->obsolete && !peer->endpoint_dst->ops->check(peer->endpoint_dst, 0))) {
-		read_unlock_bh(&peer->endpoint_lock);
-		socket_set_peer_dst(peer);
-		read_lock_bh(&peer->endpoint_lock);
-		if (!peer->endpoint_dst)
-			goto out;
-	}
-
-	if (!atomic_inc_not_zero(&peer->endpoint_dst->__refcnt))
-		goto out;
-	dst = peer->endpoint_dst;
-
-out:
-	read_unlock_bh(&peer->endpoint_lock);
-	return dst;
-}
-
-
-int socket_send_skb_to_peer(struct wireguard_peer *peer, struct sk_buff *skb, u8 ds)
-{
-	struct net_device *dev = netdev_pub(peer->device);
-	struct dst_entry *dst;
-	size_t skb_len = skb->len;
-	int ret = 0;
-
-	dst = peer_dst_get(peer);
-	if (unlikely(!dst)) {
-		net_dbg_ratelimited("No route to %pISpfsc for peer %Lu\n", &peer->endpoint_addr, peer->internal_id);
-		kfree_skb(skb);
-		return -EHOSTUNREACH;
-	} else if (unlikely(dst->dev == dev)) {
-		net_dbg_ratelimited("Avoiding routing loop to %pISpfsc for peer %Lu\n", &peer->endpoint_addr, peer->internal_id);
-		kfree_skb(skb);
-		return -ELOOP;
-	}
-
-	rcu_read_lock();
-	read_lock_bh(&peer->endpoint_lock);
-
-	ret = send(dev, skb, dst, &peer->endpoint_flow.fl4, &peer->endpoint_flow.fl6, &peer->endpoint_addr, rcu_dereference(peer->device->sock4), rcu_dereference(peer->device->sock6), ds);
-	if (!ret)
-		peer->tx_bytes += skb_len;
-
-	read_unlock_bh(&peer->endpoint_lock);
-	rcu_read_unlock();
-
-	return ret;
-}
-
-int socket_send_buffer_to_peer(struct wireguard_peer *peer, void *buffer, size_t len, u8 ds)
-{
-	struct sk_buff *skb = alloc_skb(len + SKB_HEADER_LEN, GFP_ATOMIC);
-	if (!skb)
-		return -ENOMEM;
-	skb_reserve(skb, SKB_HEADER_LEN);
-	memcpy(skb_put(skb, len), buffer, len);
-	return socket_send_skb_to_peer(peer, skb, ds);
-}
-
-static int send_to_sockaddr(struct sk_buff *skb, struct wireguard_device *wg, struct sockaddr_storage *addr, struct sock *sock4, struct sock *sock6)
-{
-	struct dst_entry *dst;
-	struct net_device *dev = netdev_pub(wg);
-	union {
-		struct flowi4 fl4;
-		struct flowi6 fl6;
-	} fl;
-
-	dst = route(wg, &fl.fl4, &fl.fl6, addr, sock4, sock6);
-	if (IS_ERR(dst)) {
-		net_dbg_ratelimited("No route to %pISpfsc\n", addr);
-		kfree_skb(skb);
-		return PTR_ERR(dst);
-	} else if (unlikely(dst->dev == netdev_pub(wg))) {
-		net_dbg_ratelimited("Avoiding routing loop to %pISpfsc\n", addr);
-		dst_release(dst);
-		kfree_skb(skb);
-		return -ELOOP;
-	}
-
-	return send(dev, skb, dst, &fl.fl4, &fl.fl6, addr, sock4, sock6, 0);
-}
-
-int socket_send_buffer_as_reply_to_skb(struct sk_buff *in_skb, void *out_buffer, size_t len, struct wireguard_device *wg)
-{
-	int ret = 0;
-	struct sk_buff *skb;
-	struct sockaddr_storage addr = { 0 };
-
-	if (unlikely(!in_skb))
-		return -EINVAL;
-	ret = socket_addr_from_skb(&addr, in_skb);
-	if (ret < 0)
-		return ret;
-
-	skb = alloc_skb(len + SKB_HEADER_LEN, GFP_ATOMIC);
-	if (!skb)
-		return -ENOMEM;
-	skb_reserve(skb, SKB_HEADER_LEN);
-	memcpy(skb_put(skb, len), out_buffer, len);
-
-	rcu_read_lock();
-	ret = send_to_sockaddr(skb, wg, &addr, rcu_dereference(wg->sock4), rcu_dereference(wg->sock6));
-	rcu_read_unlock();
-
-	return ret;
-}
-
 static int receive(struct sock *sk, struct sk_buff *skb)
 {
 	struct wireguard_device *wg;
